from typing import Any, Dict

try:
    from django.utils.decorators import classonlymethod
    from rest_framework import status
    from rest_framework.response import Response
    from rest_framework.viewsets import ViewSetMixin
except ImportError:
    print('需要安装Django以及DRF框架：pip install django djangorestframework')
    raise


class EasyViewSetMixin(ViewSetMixin):
    """
    提供了两个方法来简化 ``.as_view()`` 的传参。
    """

    @classonlymethod
    def to_view(cls, initkwargs: Dict[str, Any] = None, **actions: str):
        """
        将视图类存储在一个视图函数上，方便URL反向查找。

        之所以调转 ``initkwargs`` 和 ``actions`` 的传参方式，是因为后者往往用得更频繁。

        :param initkwargs: 视图类的初始化参数。
        :param actions: HTTP方法及其在视图类中的方法函数名称。
                        不同方法（的组合）可以让不同的视图函数应对不同的HTTP请求。
                        至少要提供一个key-value。
        :return: 一个新的视图函数。
        :raise TypeError: 来自 .as_view() 的异常。
        """
        if initkwargs is None:
            initkwargs = {}
        return cls.as_view(actions, **initkwargs)

    @classonlymethod
    def av(cls, actions: str, **initkwargs):
        """
        将视图类存储在一个视图函数上，方便URL反向查找。

        使用区分大小写的字母代表指定视图类方法及其对应的HTTP方法。

        ``actions`` 的缩写如下：
          - ``o``，对应 OPTIONS 请求及 ``.options()`` 方法。
          - ``l``，对应 GET 请求及 ``.list()`` 方法。
          - ``r``，对应 GET 请求及 ``.retrieve()`` 方法。
          - ``c``，对应 POST 请求及 ``.create()`` 方法。
          - ``u``，对应 PUT 请求及 ``.update()`` 方法。
          - ``U``，对应 PUT 请求及 ``.partial_update()`` 方法。
          - ``p``，对应 PATCH 请求及 ``.partial_update()`` 方法。
          - ``d``，对应 DELETE 请求及 ``.destroy()`` 方法。
          - ``D``，对应 DELETE 请求及 ``.soft_delete()`` 方法。

        :param actions: 缩写。相同HTTP请求的缩写不能组合，否则只择其一。
        :param initkwargs: 视图类的初始化参数。
        :return: 一个新的视图函数。
        :raise KeyError: 缩写未被定义。
        :raise TypeError: 来自 .as_view() 的异常。
        """
        # 缩写 p 是请求方法的性质向功能的妥协。
        # PATCH 请求是对资源进行部分修改，PUT 请求是对资源进行整体覆盖，与 partial 相悖；
        # 但是 PUT 请求是幂等的，而 PATCH 不是，也就是说多次
        # PATCH 相同内容产生的效果可以不一样，这与 update 相悖。
        mapper = {
            'o': ('options', 'options'),
            'l': ('get', 'list'),
            'r': ('get', 'retrieve'),
            'c': ('post', 'create'),
            'u': ('put', 'update'),
            'U': ('put', 'partial_update'),
            'p': ('patch', 'partial_update'),
            'd': ('delete', 'destroy'),
            'D': ('delete', 'soft_delete'),
        }
        return cls.as_view(dict(mapper[a] for a in actions), **initkwargs)


class SoftDeleteModelMixin:
    """
    将一个模型实例标记为已删除。（软删除）
    """
    deletion_field = 'deleted'
    deletion_mark = True

    def soft_delete(self, request, *args, **kwargs):
        instance = self.get_object()
        self.perform_soft_delete(instance)
        return Response(status=status.HTTP_204_NO_CONTENT)

    def perform_soft_delete(self, instance):
        if not hasattr(instance, self.deletion_field):
            raise TypeError(
                '模型 %s 没有用于标记删除的字段 %s 。' % (
                    type(instance).__name__,
                    self.deletion_field,
                )
            )

        setattr(instance, self.deletion_field, self.deletion_mark)

        instance.save()



"""
比较零散的不好分类的快捷函数。
"""


def true(value) -> bool:
    """
    将HTTP请求中 query 部分的参数值转换为 Python 的逻辑值。

    :param value: query 中的参数值。
    :return: True 或 False。
    """
    return value in ('true', 'True', 'TRUE', 1, True)



"""
中国行政区划相关。
"""

from .typing import IntegerChoices


class Province(IntegerChoices):
    """
    中国省级行政区。
    """
    BEIJING = 11, '北京市'
    TIANJIN = 12, '天津市'
    HEBEI = 13, '河北省'
    SHANXI = 14, '山西省'
    NEI_MONGOLIA = 15, '内蒙古自治区'
    LIAONING = 21, '辽宁省'
    JILIN = 22, '吉林省'
    HEILONGJIANG = 23, '黑龙江省'
    SHANGHAI = 31, '上海市'
    JIANGSU = 32, '江苏省'
    ZHEJIANG = 33, '浙江省'
    ANHUI = 34, '安徽省'
    FUJIAN = 35, '福建省'
    JIANGXI = 36, '江西省'
    SHANDONG = 37, '山东省'
    HENAN = 41, '河南省'
    HUBEI = 42, '湖北省'
    HUNAN = 43, '湖南省'
    GUANGDONG = 44, '广东省'
    GUANGXI = 45, '广西壮族自治区'
    HAINAN = 46, '海南省'
    CHONGQING = 50, '重庆市'
    SICHUAN = 51, '四川省'
    GUIZHOU = 52, '贵州省'
    YUNNAN = 53, '云南省'
    XIZANG = 54, '西藏自治区'
    SHAANXI = 61, '陕西省'
    GANSU = 62, '甘肃省'
    QINGHAI = 63, '青海省'
    NINGXIA = 64, '宁夏回族自治区'
    XINJIANG = 65, '新疆维吾尔自治区'
    TAIWAN = 71, '台湾省'
    HONG_KONG = 81, '香港特别行政区'
    MACAO = 82, '澳门特别行政区'


REGIONS = {
    '华东': [Province.SHANGHAI, Province.JIANGSU, Province.ZHEJIANG, Province.ANHUI, Province.JIANGXI],
    '华北': [
        Province.BEIJING, Province.TIANJIN, Province.HEBEI, Province.SHANXI, Province.NEI_MONGOLIA, Province.SHANDONG
    ],
    '华中': [Province.HENAN, Province.HUBEI, Province.HUNAN],
    '华南': [Province.FUJIAN, Province.GUANGDONG, Province.GUANGXI, Province.HAINAN],
    '东北': [Province.LIAONING, Province.JILIN, Province.HEILONGJIANG],
    '西北': [Province.SHAANXI, Province.GANSU, Province.QINGHAI, Province.NINGXIA, Province.XINJIANG],
    '西南': [Province.CHONGQING, Province.SICHUAN, Province.GUIZHOU, Province.YUNNAN, Province.XIZANG],
    '港澳台': [Province.TAIWAN, Province.HONG_KONG, Province.MACAO],
}



"""
时间相关的处理。包括日期时间类型转换和计时。
"""

import sys
from datetime import timedelta, datetime
from functools import wraps
from io import TextIOWrapper
from logging import Logger, DEBUG, getLevelName
from typing import Tuple, Optional, Union, NoReturn, TextIO


def delta2hms(delta: timedelta) -> Tuple[int, int, float]:
    """
    将时间增量转换为时分秒格式，其中秒钟以小数形式包含毫秒和微秒。

    :param delta: 时间增量。
    :return: 一个三元元组。
    """
    h = delta.seconds // 3600
    m = delta.seconds % 3600 // 60
    s = delta.seconds % 60 + delta.microseconds / 1000000
    return h, m, s


def delta2ms(delta: timedelta) -> Tuple[int, float]:
    """
    将时间增量转换为分秒格式，其中秒钟以小数形式包含毫秒和微秒。

    :param delta: 时间增量。
    :return: 二元元组。前者用一个整数表示分钟数，
             后者用一个小数表示秒钟数和纳秒数。
    """
    m = delta.seconds // 60
    s = delta.seconds % 60 + delta.microseconds / 1000000
    return m, s


def delta2s(delta: timedelta) -> float:
    """
    将时间增量转换为秒钟数，以小数形式包含毫秒和微秒。

    :param delta: 时间增量。
    :return: 一个小数。
    """
    return delta.seconds + delta.microseconds / 1000000


class BearTimer(object):
    fmt = '[{head:%H:%M:%S.%f}] [{level}] [{title}] [{total:.6f} +{delta:.6f}]: {msg}'
    level = DEBUG

    def __init__(self, title: str = None, output: Union[TextIOWrapper, TextIO, Logger] = None):
        """
        熊牌计时器。对代码运行进行计时，并打印时间和提示。

        比较简单的用法是使用 with 语句包裹需要计时的部分，
        它会在开始执行前启动计时，在执行完毕或出现异常而离开with后停止计时。

        >>> with BearTimer() as bear:
        >>>     summary = 0
        >>>     for i in range(1000):
        >>>         if not i % 67:
        >>>             bear.step(f'loop to {i} now.')
        >>>         summary += i

        如果需要对一整个函数的运行进行计时，那么可以将计时器作为装饰器使用：

        >>> @BearTimer()
        >>> def prepare_order(request, *args, **kwargs):
        >>>     # 业务逻辑
        >>>     pass

        如果不方便使用 with 包裹，可以直接实例化一个类对象。
        每一个对象都是独立的计时器，互不影响。

        >>> bear = BearTimer()
        >>> bear.start()
        >>> bear.stop()

        :param title: 计时器的标题，用以标明输出信息归属于哪个计时器。默认从打印消息时的上下文中获取。
        :param output: 消息打印到哪里去。默认是系统标准输出。可以是一个文件IO对象，也可以是一个日志记录器。
        """
        self._start = None  # 计时开始时间
        self._point = None  # 中途标记时间（用于计算距离上次标记过去了多久）
        self._title = title
        self.output = output
        if not self._title:
            if hasattr(sys, '_getframe'):
                self._title = sys._getframe().f_back.f_code.co_name
            else:
                self._title = '<UNTITLED>'

    def __call__(self, func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            self._title = func.__name__
            self.start()
            returns = func(*args, **kwargs)
            self.stop()
            return returns

        return wrapper

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, traceback):
        self.stop()

    def handle(self, full_message: str) -> NoReturn:
        """
        将消息发送到指定位置。这个位置允许是文件IO对象，或者是日志记录器。

        不应直接使用此方法。

        :param full_message: 完整的消息内容。
        :return: 无。
        """
        if hasattr(self.output, 'log') and callable(self.output.log):
            self.output: Logger
            self.output.log(self.level, full_message)

        elif hasattr(self.output, 'write') and callable(self.output.write):
            self.output: TextIOWrapper
            print(full_message, file=self.output)

        else:
            print(full_message)

    def record(self, msg: str = '', *args, **kwargs) -> datetime:
        """
        打印一行消息，标明此刻的时间、经历的时长、距上次打印的时长、计时器标题，以及附加的消息。

        不应直接使用这个方法。

        :param msg: 要附加的消息。默认为空文本。
        :param args: 其它需要打印的位置参数。
        :param kwargs: 其它需要打印的关键字参数。
        :return: 此时的时刻。
        """
        now = datetime.now()
        total = delta2s(now - self._start) if self._start else 0
        delta = delta2s(now - self._point) if self._point else 0
        title = self._title if self._title else ''
        message = self.fmt.format(
            head=now, total=total, delta=delta, title=title, msg=msg,
            level=getLevelName(self.level), *args, **kwargs,
        )
        self._point = now
        self.handle(message)
        return now

    def start(self, msg='Starting...'):
        """
        开始计时。

        如果计时尚未结束，会重新开始计时，中途记录的标记也会被清除。

        :param msg: 要附加的消息。
        :return: 计时器对象自身。
        """
        self._point = None
        self._start = datetime.now()
        self.record(msg)
        return self

    def step(self, msg='') -> Tuple[Optional[datetime], datetime]:
        """
        中途标记。

        用此方法可以计算出距离上一次标记/开始计时过去了多久。

        如果计时器尚未开始计时，那么过去的时间始终是 0 秒。

        :param msg: 要附加的消息。
        :return: 二元元组。包含上一次标记/开始计时的时刻，和此时的时刻。
        """
        prev = self._point
        curr = self.record(msg)
        return prev, curr

    def stop(self, msg='Stopped.') -> datetime:
        """
        结束计时。

        :param msg: 要附加的消息。
        :return: 此时的时刻。
        """
        curr = self.record(msg)
        self._start = None
        return curr



"""
类型相关。
"""

import datetime
import enum
import re
from types import DynamicClassAttribute
from typing import Callable, Type, Union
from uuid import UUID


class OnionObject(object):

    def __translate_list(self, items: list) -> list:
        return list(
            self.__translate_list(list(item)) if isinstance(item, tuple)
            else self.__translate_list(item) if isinstance(item, list)
            else type(self)(item, self.__depth - 1) if isinstance(item, dict)
            else item
            for item in items
        )

    # OnionObject()
    def __new__(cls, dictionary: dict = None, depth: int = -1, **kvs):
        """
        将字典转化为对象，使得可以用点分法代替下标访问内容。

        支持还原为字典：

        >>> data: dict = ~OnionObject()

        支持直接更新字典：

        >>> obj: OnionObject = OnionObject()
        >>> obj: OnionObject = obj | dict()  # 等价于 obj |= dict()

        :param dictionary: 包含数据的字典。不符合标识符命名要求，
                           或者以双下划线 “__” 开头的键不会被收录。
        :param depth: 递归转化的层数。负数表示无限递归转化，正数表示递归层数，0无意义。
        :return: OnionObject 的对象。
        :raise TypeError: 属性名称必须是字符串。
        """
        if dictionary is None:
            dictionary = {}
        if not isinstance(dictionary, dict):
            raise TypeError  # pragma: no cover
        if depth == 0:
            return dictionary
        dictionary.update(kvs)

        self = object.__new__(cls)
        self.__depth = depth
        self.__ior__(dictionary)
        return self

    # OnionObject() | dict()
    def __or__(self, dictionary: dict) -> 'OnionObject':
        for k, v in dictionary.items():
            k = str(k)
            if not k.isidentifier() or k.startswith('__'):
                continue
            if isinstance(v, tuple):
                self.__setattr__(k, self.__translate_list(list(v[:])))
            elif isinstance(v, list):
                self.__setattr__(k, self.__translate_list(v[:]))
            elif isinstance(v, dict) and self.__depth:
                self.__setattr__(k, self.__class__(v, self.__depth - 1))
            else:
                self.__setattr__(k, v)
        return self

    # obj = OnionObject()
    # obj |= dict()
    __ior__ = __or__

    # ~OnionObject()
    def __invert__(self) -> dict:
        def obtain():
            prefix = f'_{type(self).__name__}__'
            for k, v in self.__dict__.items():
                if k.startswith('__') or k.startswith(prefix):
                    continue
                if isinstance(v, (list, tuple)):
                    yield k, [
                        ~i if isinstance(i, OnionObject) else i
                        for i in v
                    ]
                elif isinstance(v, OnionObject):
                    yield k, ~v
                else:
                    yield k, v

        return dict(obtain())

    # repr(OnionObject())
    def __repr__(self) -> str:
        prefix = f'_{type(self).__name__}__'
        attrs = ', '.join(
            f'{attr}={type(value).__name__}(...)'
            if isinstance(value, OnionObject) else
            f'{attr}={value!r}'
            for attr, value in self.__dict__.items()
            if not attr.startswith('__') or attr.startswith(prefix)
        )
        return f'OnionObject({attrs})'


def casting(
        mapper: Callable,
        raw,
        *errs: Union[Exception, Type[Exception]],
        default=None,
):
    """
    使用mapper将raw转换为所需的值，当出现异常时返回default。

    默认捕获以下异常，可以通过errs参数追加更多异常：
        - TypeError
        - ValueError
        - KeyboardInterrupt

    :param mapper: 类型转换器。如果转换器不可调用，将直接返回默认值。
    :param raw: 被转换的值。
    :param errs: 需要捕获的其它异常类或异常对象。应当提供可被 except 语句接受的值。
    :param default: 默认值。即使不提供也会默认返回 None 而不会抛出异常。
    :return: 转换后的值。如若捕获到特定异常将返回默认值。
    """
    if not callable(mapper):
        return default  # pragma: no cover
    try:
        return mapper(raw)
    except (TypeError, ValueError, KeyboardInterrupt) + errs:
        return default


def represent(value) -> str:
    """
    将任意值转换为一个易于阅读的字符串。

    也是 ReprMixin 的默认格式化函数。

    默认使用 repr() 函数进行转换。如果自定义的类需要实现被此函数转换，请重写 .__repr__() 方法。

    :param value: 任意值。
    :return: 字符串。
    """
    if isinstance(value, str):
        return f'"{value}"'
    elif isinstance(value, datetime.timedelta):
        return f'[{value.days}d+{value.seconds}.{value.microseconds:06d}s]'
    elif isinstance(value, datetime.datetime):
        return f'[{value:%Y-%m-%d %H:%M:%S,%f}]'
    elif isinstance(value, datetime.date):
        return f'[{value:%Y-%m-%d}]'
    elif isinstance(value, UUID):
        return value.hex
    else:
        return repr(value)


class ReprMixin(object):
    """
    生成通用representation的工具类。

    格式类似于 ``<User(1) female name="meow" age=12 birth=[2012-01-23]>``，
    包含自身类名、主键、标签、属性名和值。

    直接或间接继承时请放在第一父类。
    """

    def __repr__(self) -> str:
        kls = self._obtain_kls()
        pkv = self._obtain_pk()
        tags = self._obtain_tags()
        attrs = self._obtain_attrs()
        content = (
                (f'({pkv})' if pkv else '') +
                (f' {tags}' if tags else '') +
                (f' {attrs}' if attrs else '')
        )
        return f'<{kls}{content}>'

    class AttributeMeta:
        """
        用于控制生成 representation 时需要带上哪些属性。

        注意：这个内部类不会被实例化！

        AttributeMeta 的变量代表你的类对象在运行时已经存在的属性，
        变量值应当是一个字符串，表示生成 representation 时这个属性的名称是什么。

        AttributeMeta 的变量允许接收一个返回值为字符串的函数作为类型注解，
        用于转换你的类对象的属性值，并直接作为 representation 里这个属性的值。
        """

    def _obtain_attrs(self) -> str:
        def obtain():
            for attr, name in attributes.items():
                if attr.startswith('_'):
                    continue
                mapper = annotations.get(attr, represent)
                value = getattr(self, attr)
                value = mapper(value) if callable(mapper) else value
                yield f'{name}={value}'

        attributes = self.AttributeMeta.__dict__
        annotations = attributes.get('__annotations__', {})
        return ' '.join(filter(None, obtain()))

    class TagMeta:
        """
        用于控制生成 representation 时需要带上哪些标签。

        注意：这个内部类不会被实例化！

        TagMeta 的变量代表你的类对象在运行时已经存在的属性，变量值可以是
          - 一个字符串，表示这个属性为 ``True`` 时 representation 里才会出现的标签的名称；
            如果属性为 ``False`` 则不会出现这个标签。
          - 一个元组且仅有两个字符串，表示这个属性分别为
            ``False`` 和 ``True`` 时 representation 里会出现的标签的名称。
          - 一个列表或一个字典，则使用属性值对这个列表或字典进行取值，
            以此作为 representation 里出现的标签的名称。

        TagMeta 的变量允许接收一个返回值为字符串的函数作为类型注解，
        用于进一步转换你的类对象的属性值，若未提供，默认使用 bool() 来转换。
        """

    def _obtain_tags(self) -> str:
        def obtain():
            for attr, option in attributes.items():
                if attr.startswith('_'):
                    continue
                mapper = annotations.get(attr, None)
                value = getattr(self, attr)
                value = mapper(value) if callable(mapper) else value
                if isinstance(option, str) and value:
                    yield option
                elif isinstance(option, tuple):
                    yield option[bool(value)]
                elif isinstance(option, (list, dict)):
                    yield option[value]
                else:
                    pass  # pragma: no cover

        attributes = self.TagMeta.__dict__
        annotations = attributes.get('__annotations__', {})
        return ' '.join(filter(None, obtain()))

    def _obtain_pk(self) -> str:
        if hasattr(self, 'pk'):
            pk = self.pk
        elif hasattr(self, 'id'):
            pk = self.id
        else:
            return ''
        return pk if isinstance(pk, str) else represent(pk)

    def _obtain_kls(self) -> str:
        return self.__class__.__name__


def datasize(literal: str) -> Union[int, float]:
    """
    将一个字面量转换为字节数目。

    支持的单位包括：
      - B、b
      - KB、KiB、Kb、Kib
      - MB、MiB、Mb、Mib
      - GB、GiB、Gb、Gib
      - TB、TiB、Tb、Tib
      - 以此类推……

    - 1 B == 8 b
    - 1 MB == 1000 KB
    - 1 MiB == 1024 KiB

    :param literal: 一个整数后缀数据大小的单位。
    :return:
    """
    if not isinstance(literal, str):
        raise TypeError(
            '不支持解析一个非字符串类型的值。'  # pragma: no cover
        )

    pattern = re.compile(r'^([0-9]+)\s*([KMGTPEZY]?)(i?[Bb])$')
    result = re.fullmatch(pattern, literal)

    if result is None:
        return 0

    base = int(result.group(1))
    shift = 'BKMGTPEZY'.index(result.group(2))
    power = (1024 if 'i' in result.group(3) else 1000) ** shift
    power = (power / 8) if 'b' in result.group(3) else power

    return base * power


# copy from https://github.com/django/django/blob/stable/4.2.x/django/db/models/enums.py
class ChoicesMeta(enum.EnumMeta):
    """用于创建带有标签文本的枚举的元类。"""

    def __new__(metacls, classname, bases, classdict, **kwds):
        labels = []
        for key in classdict._member_names:
            value = classdict[key]
            if (
                    isinstance(value, (list, tuple))
                    and len(value) > 1
                    and isinstance(value[-1], (str,))
            ):
                *value, label = value
                value = tuple(value)
            else:
                label = key.replace("_", " ").title()
            labels.append(label)
            # Use dict.__setitem__() to suppress defenses against double
            # assignment in enum's classdict.
            dict.__setitem__(classdict, key, value)
        cls = super().__new__(metacls, classname, bases, classdict, **kwds)
        for member, label in zip(cls.__members__.values(), labels):
            member._label_ = label
        return enum.unique(cls)

    def __contains__(cls, member):
        if not isinstance(member, enum.Enum):
            # Allow non-enums to match against member values.
            return any(x.value == member for x in cls)
        return super().__contains__(member)

    @property
    def names(cls):
        empty = ["__empty__"] if hasattr(cls, "__empty__") else []
        return empty + [member.name for member in cls]

    @property
    def choices(cls):
        empty = [(None, cls.__empty__)] if hasattr(cls, "__empty__") else []
        return empty + [(member.value, member.label) for member in cls]

    @property
    def labels(cls):
        return [label for _, label in cls.choices]

    @property
    def values(cls):
        return [value for value, _ in cls.choices]


# copy from https://github.com/django/django/blob/stable/4.2.x/django/db/models/enums.py
class Choices(enum.Enum, metaclass=ChoicesMeta):
    """用于创建带有标签文本的枚举的类。"""

    @DynamicClassAttribute
    def label(self):
        return self._label_

    @property
    def do_not_call_in_templates(self):
        return True

    def __str__(self):
        """
        Use value when cast to str, so that Choices set as model instance
        attributes are rendered as expected in templates and similar contexts.
        """
        return str(self.value)

    # A similar format was proposed for Python 3.10.
    def __repr__(self):
        return f"{self.__class__.__qualname__}.{self._name_}"


# copy from https://github.com/django/django/blob/stable/4.2.x/django/db/models/enums.py
class IntegerChoices(int, Choices):
    """用于创建值是整数的带有标签文本的枚举的类。"""

    pass


# copy from https://github.com/django/django/blob/stable/4.2.x/django/db/models/enums.py
class TextChoices(str, Choices):
    """用于创建值是字符串的带有标签文本的枚举的类。"""

    def _generate_next_value_(name, start, count, last_values):
        return name
